#!/usr/bin/perl

# Copyright (C) 2003 Glen Harris <astfgl@iamnota.org>,
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

use Config::IniFiles;
use GD;
use Cwd;

if ($#ARGV != 0) {
	print "usage: $0 [config.ini]\n";
	exit(0);
}

#
# Read the config. Die if it's missing.
#
if ( -f $ARGV[0] ) {
	# Try to get defaults from rc file
	$cfg = new Config::IniFiles(
		-file => $ARGV[0],
		-nocase => 1,
		-allowcontinue => 0,
		);
	if (!$cfg) {
		die "Error parsing $ARGV[0]. Aborting.";
	}
} else {
	die "File $ARGV[0] not found. Aborting.";
}

if (defined($cfg->val("global", "verbose"))) {
	$DEBUG = $cfg->val("global", "verbose");
} else {
	$DEBUG = 1;
}
print "DEBUG: verbose is $DEBUG\n" if ($DEBUG > 0);

#
# Is the output ISO filename set? Well, set it!
#
if (!defined($cfg->val("global", "isoname"))) {
	$cfg->newval("global", "isoname", "mydvd.iso");
}

#
# Let's also prepend the output ISO path to the ISO name if necessary
#
if (defined($cfg->val("global", "isodir"))) {
	$t = $cfg->val("global", "isoname");
	$cfg->delval("global", "isoname");
	$cfg->newval("global", "isoname", $cfg->val("global", "isodir")."/$t");
}

#
# Set the image size from the TV standard. Default to PAL.
#
if (defined($cfg->val("global", "format"))) {
	if ($cfg->val("global", "format") eq "PAL") {
		print "DEBUG: Setting format to PAL (720x576)\n" if ($DEBUG == 2);
		$Width = 720;
		$Height = 576;
		$cmd_ppmtoy4m .= "ppmtoy4m -n 1 -F 25:1 -A 59:54 -I t -L";
	} elsif ($cfg->val("global", "format") eq "NTSC") {
		print "DEBUG: Setting format to NTSC (720x480)\n" if ($DEBUG == 2);
		$Width = 720;
		$Height = 480;
		$cmd_ppmtoy4m .= "ppmtoy4m -n 1 -F 30000:1001 -A 10:11 -I t -L";
	} else {
		print "DEBUG: Format must be either PAL or NTSC\n" if ($DEBUG);
	}
} else {
	warn "Format not set, assuming PAL.";
	print "DEBUG: Setting format to PAL (720x576)\n" if ($DEBUG == 2);
	$Width = 720;
	$Height = 576;
	$cmd_ppmtoy4m .= "ppmtoy4m -n 1 -F 25:1 -A 59:54 -I t -L";
}

#
# Turn the font filename into an absolute path if necessary.
#
if ($cfg->val("global", "font")) {
	if (! ($cfg->val("global", "font") =~ m/^\//)) {
		print "DEBUG: Assuming font path relative, converting to absolute\n" if ($DEBUG == 2);
		$t = cwd()."/".$cfg->val("global", "font");
		print "Normalizing ".$cfg->val("global", "font")." to ".$t."\n" if ($DEBUG);
		$cfg->delval("global", "font");
		$cfg->newval("global", "font", $t);
	}
	print "DEBUG: Font file is ".$cfg->val("global", "font")."\n" if ($DEBUG == 2);
}

#
# Set missing fonts from default. Die if the default is needed but not set.
# Turn the filename into an absolute path if necessary.
#
foreach $sect ("heading", "subheading", "titles") {
	if (!$cfg->val($sect, "font")) {
		print "DEBUG: Setting $sect font from default\n" if ($DEBUG == 2);
		if (!$cfg->val("global", "font")) {
			die "If you don't set a $sect font, you MUST set a default font. Aborting.";
		}
		$cfg->newval($sect, "font", $cfg->val("global", "font"));
		print "DEBUG: $sect font is ".$cfg->val($sect, "font")."\n" if ($DEBUG == 2);
	} else {
		if (! ($cfg->val($sect, "font") =~ m/^\//)) {
			print "DEBUG: Assuming font path relative, converting to absolute\n" if ($DEBUG == 2);
			$t = cwd()."/".$cfg->val($sect, "font");
			print "Normalizing ".$cfg->val($sect, "font")." to ".$t."\n" if ($DEBUG);
			$cfg->delval($sect, "font");
			$cfg->newval($sect, "font", $t);
		}
		print "DEBUG: Font file is ".$cfg->getval($sect, "font")."\n" if ($DEBUG == 2);
	}
}

#
# GD renders TTF better in a truecolor image. We turn one image back to
# indexed later. Antialiasing in turned OFF for BOTH, however.
#
GD::Image->trueColor(1);

#
# Parse the text color.
#
($R, $G, $B) = split(/ /, $cfg->val("global", "textcolor"));
print "DEBUG: Text color is $R $G $B\n" if ($DEBUG == 2);

#
# Create the background image and allocate the text drawing color.
#
if ($cfg->val("global", "background")) {
	print "DEBUG: Background image is ".$cfg->val("global", "background")."\n" if ($DEBUG == 2);
	if ( -f $cfg->val("global", "background") ) {
		$img_b = GD::Image->new($cfg->val("global", "background"));
		if (!defined($img_b)) {
			die "libgd2 could not read image ".$cfg->val("global", "background").". Aborting.";
		}
		($w, $h) = $img_b->getBounds();
		if (($w != $Width) || ($h != $Height)) {
			die "Background image is ".$w."x".$h.", should be ".$Width."x".$Height.". Aborting.";
		}
	} else {
		die "Cannot open background image file ".$cfg->val("global", "background").". Aborting.";
	}
} else {
	print "DEBUG: Creating black background image\n" if ($DEBUG == 2);
	$img_b = new GD::Image($Width, $Height);
}
$img_b_textcolor = $img_b->colorAllocate($R, $G, $B);

#
# Create the mask image image and allocate black, white and text colors.
# White will be transparent, but we don't set it until after turning the
# image back to indexed.
#
print "DEBUG: Creating mask layer\n" if ($DEBUG == 2);
$img_m = new GD::Image($Width, $Height);
$img_m_white = $img_m->colorAllocate(255,255,255);
$img_m_black = $img_m->colorAllocate(0,0,0);
$img_m_textcolor = $img_m->colorAllocate($R, $G, $B);
$img_m->fill(1,1,$img_m_white);

$YPOS_HEADER	= 0;
$YPOS_1ST_TITLE = 0;

#
# Start at the top of the screen
#
$YPOS = 0;

#
# Process the HEADING section if it exists
#
if ($cfg->SectionExists("heading")) {
	print "DEBUG: Found a HEADING section\n" if ($DEBUG == 2);
	if ($cfg->val("heading", "yoffset")) {
		$YPOS += $cfg->val("heading", "yoffset");
	} else {
		$YPOS += 10;
	}
	
	$YPOS_HEADER=$YPOS;

	#
	# Place the Heading text on the background image.
	#
	@bounds = GD::Image->stringFT(0-$img_b_textcolor,
		$cfg->val("heading", "font"),
		$cfg->val("heading", "size"),
		0, 0, 0,
		$cfg->val("heading", "text"));
	$tw = $bounds[4]-$bounds[0];
	$th = $bounds[1]-$bounds[5];
	$YPOS += $th;
	$x = ($Width - $tw) / 2;
	$y = $YPOS;
	$img_b->stringFT(0-$img_b_textcolor,
		$cfg->val("heading", "font"),
		$cfg->val("heading", "size"), 0, $x, $y,
		$cfg->val("heading", "text"));
	print "Wrote heading text \"".$cfg->val("heading", "text").
		"\" (".$tw."x".$th.") at $x,$y\n" if ($DEBUG);
}

#
# Process the SUBHEADING section if it exists
#
if ($cfg->SectionExists("subheading")) {
	print "DEBUG: Found a SUBHEADING section\n" if ($DEBUG == 2);
	if ($cfg->val("subheading", "yoffset")) {
		$YPOS += $cfg->val("subheading", "yoffset");
	} else {
		$YPOS += 20;
	}
	
	#
	# Place the SubHeading text on the background image.
	#
	@bounds = GD::Image->stringFT(0-$textcolor,
		$cfg->val("subheading", "font"),
		$cfg->val("subheading", "size"), 0, 0, 0,
		$cfg->val("subheading", "text"));
	$tw = $bounds[4]-$bounds[0];
	$th = $bounds[1]-$bounds[5];
	$YPOS += $th;
	$x = ($Width - $tw) / 2;
	$y = $YPOS;
	$img_b->stringFT(0-$img_b_textcolor,
		$cfg->val("subheading", "font"),
		$cfg->val("subheading", "size"), 0, $x, $y,
		$cfg->val("subheading", "text"));
	print "Wrote subheading text \"".$cfg->val("subheading", "text").
		"\" (".$tw."x".$th.") at $x,$y\n" if ($DEBUG);
}

#
# Loop through each title found in the config file and check it.
#
$loop = 1;
while ($cfg->SectionExists("title$loop")) {
	print "DEBUG: Found TITLE section: $loop\n" if ($DEBUG == 2);
	print "DEBUG: Files in this title:\n" if ($DEBUG == 2);
	if (ref($cfg->val("title$loop", "file")) eq "SCALAR") {
		if (! -f $cfg->val("title$loop", "file")) {
			die "The file ".$cfg->val("title$loop", "file")." does not exist";
		} else {
			print "DEBUG: ".$cfg->val("title$loop", "file")." OK\n" if ($DEBUG == 2);
		}
	} else {
		@files = $cfg->val("title$loop", "file");
		foreach $t (@files) {
			if (! -f $t) {
				die "The file $t does not exist";
			} else {
				print "DEBUG: $t OK\n" if ($DEBUG == 2);
			}
		}
	}
	$loop++;
}
$NumTitles = $loop-1;
print "Found $NumTitles titles in config file.\n" if ($DEBUG);

#
# Work out how high each title line will be.
#
$YPOS +=$cfg->val("titles", "gap");
$YPOS_1ST_TITLE = $YPOS;
@bounds = GD::Image->stringFT(-$textcolor,
	$cfg->val("titles", "font"),
	$cfg->val("titles", "size"),
	0, 0, 0,
	"TgyP");
$TitleHeight = $bounds[1] - $bounds[5];

#
# Will the titles fit on the page?
#
if (($YPOS + $TitleHeight * $NumTitles) >= $Height) {
	warn "Oops, that many titles won't fit on the screen!";
}

$bcmd = "";
$tcmd = "";

$cmds_audio = "";

#
# Done checking, draw the text for each title.
#
$loop = 1;
while ($loop <= $NumTitles) {
	# Work out the line size and start position.
	@bounds = GD::Image->stringFT(-$textcolor,
		$cfg->val("titles", "font"),
		$cfg->val("titles", "size"),
		0, 0, 0,
		$cfg->val("title$loop", "text"));
	$tw = $bounds[4]-$bounds[0];
	$th = $bounds[1]-$bounds[5];
	$YPOS += $cfg->val("titles", "yoffset");
	$YPOS += $TitleHeight;
	$x = $cfg->val("titles", "xoffset");
	$y = $YPOS;

	# Draw the text on the background image.
	@bounds = $img_b->stringFT(-$img_b_textcolor,
		$cfg->val("titles", "font"),
		$cfg->val("titles", "size"),
		0, $x, $y,
		$cfg->val("title$loop", "text"));
	print "Wrote text \"".$cfg->val("title$loop", "text").
		"\" (".$tw."x".$th.") at $x,$y\n" if ($DEBUG);

	# Draw the rectangle on the mask image.
	$img_m->filledRectangle($bounds[6]-5,$bounds[7]-5,
		$bounds[2]+3,$bounds[7]+$TitleHeight+3,
		$img_m_black);

	# Draw the text on the mask image.
	$img_m->stringFT(-$img_m_textcolor, 
	 	$cfg->val("titles", "font"),
		$cfg->val("titles", "size"),
		0, $x, $y,
		$cfg->val("title$loop", "text"));
	
	$AudioTrackNumber = $cfg->val("title$loop", "audiotracknumber");

	# Cache the lines for the dvdauthor command line
	$bcmd .= "  -b ";
	$bcmd .= ($bounds[6]-5)."x".($bounds[7]-5);
	$bcmd .= "-";
	$bcmd .= ($bounds[2]+3)."x".($bounds[7]+$TitleHeight+3);

	if ($AudioTrackNumber>1) 
	{
		$bcmd .= ",vtsm$loop";
	} else {
		$bcmd .= ",vts$loop";
	}

	#
	# AUDIO MENU START
	#
	$acmd = "";

	if ($AudioTrackNumber>1) 
	{
		# cache the lines for the dvdauthor command line
		$acmd .= " -m ";

		print "DEBUG: Creating audio black background image\n" if ($DEBUG == 2);
		$img_ab = new GD::Image($Width, $Height);
		$img_ab_textcolor = $img_ab->colorAllocate($R, $G, $B);

		print "DEBUG: Creating audio mask layer\n" if ($DEBUG == 2);
		$img_a = new GD::Image($Width, $Height);
		$img_a_white = $img_a->colorAllocate(255,255,255);
		$img_a_black = $img_a->colorAllocate(0,0,0);
		$img_a_textcolor = $img_a->colorAllocate($R, $G, $B);
		$img_a->fill(1,1,$img_a_white);

		$YPOS_AT = $YPOS_HEADER;
		# Work out the line size and start position.
		@bounds = GD::Image->stringFT(-$textcolor,
			$cfg->val("heading", "font"),
			$cfg->val("heading", "size"),
			0, 0, 0,
			$cfg->val("title$loop", "text"));
		$tw = $bounds[4]-$bounds[0];
		$th = $bounds[1]-$bounds[5];
		$YPOS_AT += $th;
		$x = ($Width - $tw) / 2;
		$y = $YPOS_AT;

		# Draw the text on the background image.
		$img_ab->stringFT(0-$img_ab_textcolor,
			$cfg->val("heading", "font"),
			$cfg->val("heading", "size"), 0, $x, $y,
			$cfg->val("title$loop", "text"));

		if ($cfg->val("subheading", "yoffset")) {
			$YPOS_AT += $cfg->val("subheading", "yoffset");
		} else {
			$YPOS_AT += 20;
		}
		@bounds = GD::Image->stringFT(0-$textcolor,
			$cfg->val("subheading", "font"),
			$cfg->val("subheading", "size"), 0, 0, 0,
			"audio");
		$tw = $bounds[4]-$bounds[0];
		$th = $bounds[1]-$bounds[5];
		$YPOS_AT += $th;
		$x = ($Width - $tw) / 2;
		$y = $YPOS_AT;
		$img_ab->stringFT(0-$img_ab_textcolor,
			$cfg->val("subheading", "font"),
			$cfg->val("subheading", "size"), 0, $x, $y,
			"audio");

		$loop_a = 0;
		$YPOS_AT = $YPOS_1ST_TITLE;
		while ($loop_a < $AudioTrackNumber) {
			# Work out the line size and start position.
			$AudioTrackType = $cfg->val("title$loop", "audiotracktype$loop_a");
			@bounds = GD::Image->stringFT(-$textcolor,
				$cfg->val("titles", "font"),
				$cfg->val("titles", "size"),
				0, 0, 0,
				"audio ch. $loop_a  $AudioTrackType");
			$tw = $bounds[4]-$bounds[0];
			$th = $bounds[1]-$bounds[5];
			$YPOS_AT += $cfg->val("titles", "yoffset");
			$YPOS_AT += $TitleHeight;
			$x = $cfg->val("titles", "xoffset");
			$y = $YPOS_AT;

			# Draw the text on the background image.
			@bounds = $img_ab->stringFT(-$img_ab_textcolor,
				$cfg->val("titles", "font"),
				$cfg->val("titles", "size"),
				0, $x, $y,
				"audio ch. $loop_a  $AudioTrackType");
			print "Wrote text \"audio ch. $loop_a  $AudioTrackType".
				"\" (".$tw."x".$th.") at $x,$y\n" if ($DEBUG);

			# Draw the rectangle on the mask image.
			$img_a->filledRectangle($bounds[6]-5,$bounds[7]-5,
				$bounds[2]+3,$bounds[7]+$TitleHeight+3,
				$img_a_black);

			# Draw the text on the mask image.
			$img_a->stringFT(-$img_a_textcolor, 
				$cfg->val("titles", "font"),
				$cfg->val("titles", "size"),
				0, $x, $y,
				"audio ch. $loop_a  $AudioTrackType");

			# cache the lines for the dvdauthor command line
			$acmd .= "  -b ";
			$acmd .= ($bounds[6]-5)."x".($bounds[7]-5);
			$acmd .= "-";
			$acmd .= ($bounds[2]+3)."x".($bounds[7]+$TitleHeight+3);
			$acmd .= ",audio$loop_a+vts1";

			$loop_a++;
		}

		$img_a->trueColorToPalette(0, 4);
		$img_a->transparent($img_a->getPixel(1,1));
		open(OUT, ">menumask_audio_title$loop.png") || die;
		print OUT $img_a->png(0);
		close OUT;

		open(OUT, ">menu_audio_title$loop.jpg") || die;
		print OUT $img_ab->jpeg(100);
		close OUT;

		open(OUT, ">menu_audio_title$loop.sub") || die;
		print OUT "menumask_audio_title$loop.png 00:00:00.00 00:00:00.00 0 0 0 0 255 255\n";
		close OUT;

		$cmds_audio .= "jpegtopnm menu_audio_title$loop.jpg | ".$cmd_ppmtoy4m." | \\\n";
		$cmds_audio .= "  mpeg2enc ";
		if ($cfg->val("global", "format") eq "PAL") {
			$cmds_audio .= "-n p ";
		} else {
			$cmds_audio .= "-n n ";
		}
		$cmds_audio .= " -f 8 -o menu_audio_title$loop.m2v\n";
		$cmds_audio .= "mplex -f 8 -o /dev/stdout menu_audio_title$loop.m2v silence.mp2 | \\\n";
		$cmds_audio .= "  submux menu_audio_title$loop.sub > menu_audio_title$loop.mpg\n";

		$acmd .= " menu_audio_title$loop.mpg --entry=audio";
	}

	#
	# AUDIO MENU END
	#

	$chapters = "-c ".$cfg->val("title$loop", "chapters");

	if (ref($cfg->val("title$loop", "file")) eq "SCALAR") {
		$tcmd .= "dvdauthor -o ".$cfg->val("global", "dvddir")." ".$acmd." -p xste-palette.yuv -t ".$chapters." ".$cfg->val("title$loop", "file")." -i post=vmgm1\n";
	} else {
		$tcmd .= "dvdauthor -o ".$cfg->val("global", "dvddir")." ".$acmd." -p xste-palette.yuv -t";
		$chloop = 1;
		@files = $cfg->val("title$loop", "file");
		foreach(@files) {
			$tcmd .= " ".$chapters." $_";
			$chloop++;
		}
		$bcmd .= " \\\n";
		$tcmd .= " -i post=vmgm1\n";
	}
	
	$loop++;
}

$img_m->trueColorToPalette(0, 4);
$img_m->transparent($img_m->getPixel(1,1));
open(OUT, ">menumask.png") || die;
print OUT $img_m->png(0);
close OUT;

open(OUT, ">menu.jpg") || die;
print OUT $img_b->jpeg(100);
close OUT;

open(OUT, ">menu.sub") || die;
print OUT "menumask.png 00:00:00.00 00:00:00.00 0 0 0 0 255 255\n";
close OUT;


#
# Here's where we assemble the commands which need to be run. Make them a
# single string to be either printed out or written to a file.
#
$cmds = "";
$cmds .= "rm -rf ".$cfg->val("global", "isoname")." ".$cfg->val("global", "dvddir")."\n";

$cmds .= $cmds_audio;

$cmds .= "jpegtopnm menu.jpg | ".$cmd_ppmtoy4m." | \\\n";

$cmds .= "  mpeg2enc ";
if ($cfg->val("global", "format") eq "PAL") {
	$cmds .= "-n p ";
} else {
	$cmds .= "-n n ";
}
$cmds .= " -f 8 -o menu.m2v\n";

$cmds .= "mplex -f 8 -o /dev/stdout menu.m2v silence.mp2 | \\\n";
$cmds .= "  submux menu.sub > menu.mpg\n";

$cmds .= "dvddirgen -o ".$cfg->val("global", "dvddir")." -r\n";

$cmds .= $tcmd;

$cmds .= "dvdauthor -o ".$cfg->val("global", "dvddir")." ";
$cmds .= "-T -m \\\n";
$cmds .= $bcmd;
#print "  -p palette.yuv \\\n"; # TODO
$cmds .= "  menu.mpg -p xste-palette.yuv\n";
$cmds .= "mkisofs -dvd-video -o ".$cfg->val("global", "isoname")." ".$cfg->val("global", "dvddir")."\n";

if (defined($cfg->val("global", "scriptname"))) {
	open(OUTF, ">".$cfg->val("global", "scriptname")) or \
		die "Cannot open ".$cfg->val("global", "scriptname")." for write";
	print OUTF "#!/bin/bash\n\n";
	print OUTF $cmds;
	close(OUTF);
	chmod(0755,  $cfg->val("global", "scriptname"));
} else {
	print $cmds;
}

